from datetime import time, timedelta
import itertools
import random
import datetime
from django.utils import timezone as tz
import pytz

import pytest
from model_bakery import baker

from aircox import models


@pytest.fixture
def stations():
    return baker.make(models.Station, _quantity=2)


@pytest.fixture
def programs(stations):
    items = list(
        itertools.chain(
            *(
                baker.make(
                    models.Program, station=station, cover=None, _quantity=2
                )
                for station in stations
            )
        )
    )
    for item in items:
        item.save()
    return items


@pytest.fixture
def program(programs):
    return programs[0]


@pytest.fixture
def sched_initials(programs):
    # use concrete class; timezone is provided in order to ensure DST
    items = [
        baker.prepare(
            models.Schedule,
            program=program,
            time=time(16, 00),
            timezone="Europe/Brussels",
        )
        for program in programs
    ]
    models.Schedule.objects.bulk_create(items)
    return items


@pytest.fixture
def sched_reruns(sched_initials):
    # use concrete class
    items = [
        baker.prepare(
            models.Schedule,
            initial=initial,
            program=initial.program,
            date=initial.date,
            time=(initial.start + timedelta(hours=1)).time(),
        )
        for initial in sched_initials
    ]
    models.Schedule.objects.bulk_create(items)
    return items


@pytest.fixture
def schedules(sched_initials, sched_reruns):
    return sched_initials + sched_reruns


@pytest.fixture
def episodes(programs):
    return [
        baker.make(models.Episode, parent=program, cover=None)
        for program in programs
    ]


@pytest.fixture
def episode(episodes):
    return episodes[0]


@pytest.fixture
def podcasts(episodes):
    items = []
    for episode in episodes:
        sounds = baker.prepare(
            models.Sound,
            episode=episode,
            program=episode.program,
            is_public=True,
            _quantity=2,
        )
        for i, sound in enumerate(sounds):
            sound.file = f"test_sound_{episode.pk}_{i}.mp3"
        items += sounds
    return items


@pytest.fixture
def diff_initials(episodes):
    items = [
        baker.make(
        models.Diffusion,
        episode=episode,
        program=episode.program,
        start=datetime.datetime.today().replace(hour=12, minute=0, second=0, microsecond=0) + timedelta(minutes=30 * i),
        end=datetime.datetime.today().replace(hour=12, minute=30, second=0, microsecond=0) + timedelta(minutes=30 * i),
        )
        for i, episode in enumerate(episodes)
    ]
    return items

@pytest.fixture
def diff_reruns(diff_initials):
    items = [
        baker.make(
        models.Diffusion,
        start=initial.start + timedelta(days=1),
        end=initial.start + timedelta(days=1),
        episode=initial.episode,
        program=initial.program,
        initial=initial,
        )
        for initial in diff_initials[0:4]
    ]
    return items


@pytest.fixture
def diffusions(diff_initials, diff_reruns):
    return diff_initials + diff_reruns

@pytest.fixture
def diff_initial(diff_initials):
    return diff_initials[0]

@pytest.fixture
def diff_rerun(diff_reruns):
    return diff_reruns[0]


@pytest.fixture
def diff_time_var(episodes):
    random_episode = random.choice([episodes[0], episodes[1], episodes[2]])
    now_date = datetime.date.today()
    now_datetime = tz.now()
    tomorrow_datetime = now_datetime + timedelta(hours=25)
    yesterday_datetime = now_datetime - timedelta(hours=25)
    now_diff = baker.make(
        models.Diffusion,
        episode=random_episode,
        start=now_datetime - timedelta(minutes=30),
        end=now_datetime + timedelta(minutes=30),
    )
    after_diff = baker.make(
        models.Diffusion,
        episode=random_episode,
        start=now_datetime + timedelta(minutes=30),
        end=now_datetime + timedelta(minutes=60),
    )
    tomorrow_diff = baker.make(
        models.Diffusion,
        episode=random_episode,
        start=tomorrow_datetime,
        end=tomorrow_datetime + timedelta(minutes=30),
    )
    before_diff = baker.make(
        models.Diffusion,
        episode=random_episode,
        start=now_datetime - timedelta(minutes=60),
        end=now_datetime - timedelta(minutes=30),
    )
    yesterday_diff = baker.make(
        models.Diffusion,
        episode=random_episode,
        start=yesterday_datetime,
        end=yesterday_datetime + timedelta(minutes=30),
    )
    return (
        now_date,
        now_datetime,
        yesterday_diff,
        before_diff,
        now_diff,
        after_diff,
        tomorrow_diff,
    )
