import pytest
from model_bakery import baker

import random
import datetime
from django.utils import timezone as tz
from datetime import timedelta
from django.utils.timezone import make_aware
import pytz

from aircox.models import Diffusion


class TestDiffusionQuerySet:
    @pytest.mark.django_db
    def test_episode_by_obj(self, episodes, diffusions):
        diff_with_key_episode, diff_without_key_episode = [], []
        for diffusion in diffusions:
            if diffusion.episode is episodes[0]:
                diff_with_key_episode.append(diffusion)
            else:
                diff_without_key_episode.append(diffusion)        
        result = Diffusion.objects.episode(episode=episodes[0])
        
        assert all(diff in result for diff in diff_with_key_episode)
        assert all(diff not in result for diff in diff_without_key_episode)

    @pytest.mark.django_db
    def test_episode_by_id(self, episodes, diffusions):
        diff_with_key_episode, diff_without_key_episode = [], []
        for diffusion in diffusions:
            if diffusion.episode is episodes[0]:
                diff_with_key_episode.append(diffusion)
            else:
                diff_without_key_episode.append(diffusion) 
        result = Diffusion.objects.episode(id=episodes[0].id)

        assert all(diff in result for diff in diff_with_key_episode)
        assert all(diff not in result for diff in diff_without_key_episode)

    @pytest.mark.django_db
    def test_on_air(self, diffusions):
        assert all(
            diff.type == Diffusion.TYPE_ON_AIR
            for diff in Diffusion.objects.on_air()
        )

    # TODO test si il y a doublon qu'un seul objet soit récupéré.
    @pytest.mark.django_db
    def test_now_test(self, diff_time_var):
        now_diff, result = diff_time_var[4], Diffusion.objects.now(
            now=tz.now(), order=True
        )
        assert (
            len(result) == 1
            and result[0] == now_diff
            and now_diff.start <= tz.now() <= now_diff.end
        )

    @pytest.mark.django_db
    def test_date_without_date_arg(self, diff_time_var):
        before_diff, now_diff, after_diff = diff_time_var[3:6]
        result = Diffusion.objects.date()
        assert len(result) == 3
        assert set([before_diff, now_diff, after_diff]).issubset(set(result))

    @pytest.mark.django_db
    def test_date_with_date_arg(self, diff_time_var):
        tomorrow_date = datetime.date.today() + timedelta(days=1)
        tomorrow_diff = diff_time_var[6]
        assert Diffusion.objects.date(date=tomorrow_date).count() == 1
        assert Diffusion.objects.date(date=tomorrow_date)[0] == tomorrow_diff

    @pytest.mark.django_db
    def test_at_datetimeobj(self, diff_time_var):
        now_datetime, now_diff = diff_time_var[1], diff_time_var[4]
        assert Diffusion.objects.at(date=now_datetime).count() == 1
        assert Diffusion.objects.at(date=now_datetime)[0] == now_diff

    @pytest.mark.django_db
    def test_at_dateobj(self, diff_time_var):
        now_date = diff_time_var[0]
        before_diff, now_diff, after_diff = diff_time_var[3:6]
        result = Diffusion.objects.at(date=now_date)
        assert len(result) == 3
        assert set([before_diff, now_diff, after_diff]).issubset(set(result))

    @pytest.mark.django_db
    def test_after_without_arg(self, diff_time_var):
        before_diff, now_diff, after_diff = diff_time_var[3:6]
        tomorrow_diff, yesterday_diff = diff_time_var[6], diff_time_var[2]
        from_now_diff = Diffusion.objects.after()
        assert set([now_diff, after_diff, tomorrow_diff]).issubset(
            set(from_now_diff)
        )
        assert (
            before_diff not in from_now_diff
            and yesterday_diff not in from_now_diff
        )

    @pytest.mark.django_db
    def test_after_with_datetime_arg(self, diff_time_var):
        now_datetime = diff_time_var[1]
        before_diff, now_diff, after_diff = diff_time_var[3:6]
        tomorrow_diff, yesterday_diff = diff_time_var[6], diff_time_var[2]
        from_now_diff = Diffusion.objects.after(date=now_datetime)
        assert set([now_diff, after_diff, tomorrow_diff]).issubset(
            set(from_now_diff)
        )
        assert (
            before_diff not in from_now_diff
            and yesterday_diff not in from_now_diff
        )

    @pytest.mark.django_db
    def test_after_with_date_arg(self, diff_time_var):
        now_date = diff_time_var[0]
        before_diff, now_diff, after_diff = diff_time_var[3:6]
        tomorrow_diff, yesterday_diff = diff_time_var[6], diff_time_var[2]
        from_today_diff = Diffusion.objects.after(date=now_date)
        assert set(
            [tomorrow_diff, after_diff, now_diff, before_diff]
        ).issubset(set(from_today_diff))
        assert yesterday_diff not in from_today_diff

    @pytest.mark.django_db
    def test_before_without_arg(self, diff_time_var):
        before_diff, now_diff, after_diff = diff_time_var[3:6]
        tomorrow_diff, yesterday_diff = diff_time_var[6], diff_time_var[2]
        before_now_diff = Diffusion.objects.before()
        assert set([now_diff, before_diff, yesterday_diff]).issubset(
            set(before_now_diff)
        )
        assert (
            tomorrow_diff not in before_now_diff
            and after_diff not in before_now_diff
        )

    @pytest.mark.django_db
    def test_before_with_datetime_arg(self, diff_time_var):
        now_datetime = diff_time_var[1]
        before_diff, now_diff, after_diff = diff_time_var[3:6]
        tomorrow_diff, yesterday_diff = diff_time_var[6], diff_time_var[2]
        before_now_diff = Diffusion.objects.before(date=now_datetime)
        assert (
            tomorrow_diff not in before_now_diff
            and after_diff not in before_now_diff
        )
        assert set([now_diff, before_diff, yesterday_diff]).issubset(
            set(before_now_diff)
        )

    @pytest.mark.django_db
    def test_before_with_date_arg(self, diff_time_var):
        now_date = diff_time_var[0]
        now_diff, after_diff = diff_time_var[4:6]
        tomorrow_diff, yesterday_diff = diff_time_var[6], diff_time_var[2]
        before_today_diff = Diffusion.objects.before(date=now_date)
        assert not set([tomorrow_diff, after_diff, now_diff]).issubset(
            set(before_today_diff)
        )
        assert yesterday_diff in before_today_diff
    
    @pytest.mark.django_db
    def test_range(self, episodes, diff_time_var):
        before_diff, now_diff, after_diff = diff_time_var[3:6]
        tomorrow_diff, yesterday_diff = diff_time_var[6], diff_time_var[2]
        start, end = before_diff.start, after_diff.end
        overlapping_diffs = [
            baker.make(
                Diffusion,
                episode=episodes[0],
                start=start - timedelta(minutes=30) if i == 0 else start,
                end=end + timedelta(minutes=30) if i == 1 else end,
            ) for i in range(2)
        ]
        range_diff = Diffusion.objects.range(start=start, end=end)
        assert {before_diff, now_diff, after_diff}.issubset(range_diff)
        assert not set(overlapping_diffs + [tomorrow_diff, yesterday_diff]).intersection(range_diff)


class TestDiffusion:
    @pytest.mark.django_db
    def test__str__initial_diff(self, diff_initial):
        diff_initial.episode.title = "Episode title"
        diff_initial.start = datetime.datetime(
            2023, 5, 12, 13, 0, tzinfo=pytz.UTC
        )
        assert diff_initial.__str__() == "Episode title - 2023/05/12 13:00+0000"

    @pytest.mark.django_db
    def test__str__rerun_diff(self, diff_initial, diff_rerun):
        diff_initial.episode.title = "Episode title"
        diff_rerun.start = datetime.datetime(
            2023, 5, 12, 13, 0, tzinfo=pytz.UTC
        )
        assert (
            diff_rerun.__str__()
            == "Episode title - 2023/05/12 13:00+0000 (rerun)"
        )

    """ je ne comprends pas dans la méthode save() le rerun_set.update, que la méthode l'update ou pas, on accède quand même à l'épisode mis à jour sur la diff initiale depuis la diff rerun"""
    @pytest.mark.django_db
    def test_save(self, episodes, diff_initial, diff_rerun):
        diff_initial.episode = episodes[1]
        diff_initial.program = episodes[1].program
        diff_initial.save()
        assert diff_rerun.initial.episode == episodes[1]
        assert diff_rerun.initial.program == episodes[1].program

    @pytest.mark.django_db
    def test_save_rerun(self, diff_initial, diff_rerun):
        diff_rerun.save_rerun()
        assert diff_rerun.episode == diff_initial.episode

    @pytest.mark.django_db
    def test_save_initial(self, diff_initial):
        diff_initial.save_initial()
        assert diff_initial.program == diff_initial.episode.program

    @pytest.mark.django_db
    def test_duration(self, diffusions):
        assert diffusions[0].duration == datetime.timedelta(seconds=1800)

    @pytest.mark.django_db
    def test_date(self, diffusions):
        diffusions[0].start = datetime.datetime(2023, 5, 12, 13, 0, tzinfo=pytz.UTC)
        diffusions[0].end = datetime.datetime(2023, 5, 12, 13, 30, tzinfo=pytz.UTC)
        assert diffusions[0].date == datetime.date(2023, 5, 12)

    @pytest.mark.django_db
    def test_local_start(self, diffusions):
        diffusions[0].start = datetime.datetime(2023, 5, 12, 13, 0, tzinfo=pytz.UTC)
        diffusions[0].end = datetime.datetime(2023, 5, 12, 13, 30, tzinfo=pytz.UTC)
        assert diffusions[0].local_start == datetime.datetime(
            2023, 5, 12, 13, 00, tzinfo=pytz.UTC
        )

    @pytest.mark.django_db
    def test_local_end(self, diffusions):
        diffusions[0].start = datetime.datetime(2023, 5, 12, 13, 0, tzinfo=pytz.UTC)
        diffusions[0].end = datetime.datetime(2023, 5, 12, 13, 30, tzinfo=pytz.UTC)
        assert diffusions[0].local_end == datetime.datetime(
            2023, 5, 12, 13, 30, tzinfo=pytz.UTC
        )

    @pytest.mark.django_db
    def test_is_now(self, diffusions):
        diffusions[0].type = Diffusion.TYPE_ON_AIR
        diffusions[0].start = tz.now()
        diffusions[0].end = tz.now() + timedelta(minutes=30)
        assert diffusions[0].is_now

    @pytest.mark.django_db
    def test_is_live(self, diffusions):
        diffusions[0].type = Diffusion.TYPE_ON_AIR
        diffusions[0].episode.sound_set == None
        assert diffusions[0].is_live

    """ pas réussi à créer les sons de type "archive" nécessaire à la méthode get_playlist """
    # @pytest.mark.django_db
    # def test_get_playlist(self, diff_initial):

    # Use fixtures podcasts as test sound file.
    @pytest.mark.django_db
    def test_get_sound(self, diffusions, podcasts):
        for podcast in podcasts:
            podcast.save()
            diffusions[0].episode.sound_set.add(podcast)
        diffusions[0].save()

        assert diffusions[0].get_sounds(removed=True).count() == 8
        assert podcasts[0] in diffusions[0].get_sounds(removed=True)

    @pytest.mark.django_db
    def test_is_date_in_range_with_arg(self, diffusions):
        duration = (diffusions[0].end - diffusions[0].start).total_seconds()
        random_time_in_range = diffusions[0].start + timedelta(
            seconds=int(random.uniform(0, duration))
        )
        assert diffusions[0].is_date_in_range(date=random_time_in_range)

    @pytest.mark.django_db
    def test_is_date_in_range_without_arg(self, diffusions):
        diffusions[0].start = make_aware(
            datetime.datetime.now() + datetime.timedelta(hours=1)
        )
        diffusions[0].end = make_aware(
            datetime.datetime.now() + datetime.timedelta(hours=2)
        )
        assert not diffusions[0].is_date_in_range()

    @pytest.mark.django_db
    def test_get_conflicts_overlapping_diffusions(
        self, diff_time_var, episodes
    ):
        new_diff = baker.make(
            Diffusion,
            start=diff_time_var[4].end - timedelta(minutes=10),
            end=diff_time_var[4].end + timedelta(hours=1),
            episode=episodes[0],
        )
        conflicts = new_diff.get_conflicts()
        assert conflicts.count() == 2
        assert all(
            x in conflicts for x in [diff_time_var[4], diff_time_var[5]]
        )

    @pytest.mark.django_db
    def test_get_conflicts_no_conflict(self, diff_time_var):
        assert diff_time_var[4].get_conflicts().count() == 0

    @pytest.mark.django_db
    def test_check_conflicts_new_conflicts(self, diff_time_var, episodes):
        new_diff = baker.make(
            Diffusion,
            start=diff_time_var[4].end - timedelta(minutes=10),
            end=diff_time_var[4].end + timedelta(hours=1),
            episode=episodes[0],
        )
        new_diff.check_conflicts()
        assert new_diff.conflicts.count() == 2
        assert all(
            x in new_diff.conflicts
            for x in [diff_time_var[4], diff_time_var[5]]
        )

    @pytest.mark.django_db
    def test_check_conflicts_additional_conflicts(
        self, diff_time_var, episodes
    ):
        new_diff = baker.make(
            Diffusion,
            start=diff_time_var[4].end - timedelta(minutes=10),
            end=diff_time_var[4].end + timedelta(hours=1),
            episode=episodes[0],
        )
        new_diff.check_conflicts()
        new_diff.start = diff_time_var[3].end - timedelta(minutes=10)
        new_diff.end = diff_time_var[3].end + timedelta(hours=1)
        new_diff.save()
        new_diff.check_conflicts()
        assert new_diff.conflicts.count() == 2
        assert all(
            x in new_diff.conflicts
            for x in [diff_time_var[3], diff_time_var[4]]
        )

    @pytest.mark.django_db
    def test_check_conflicts_remove_conflicts(self, diff_time_var, episodes):
        new_diff = baker.make(
            Diffusion,
            start=diff_time_var[4].end - timedelta(minutes=10),
            end=diff_time_var[4].end + timedelta(hours=1),
            episode=episodes[0],
        )
        new_diff.check_conflicts()
        new_diff.start = diff_time_var[5].end
        new_diff.end = diff_time_var[5].end + timedelta(minutes=30)
        new_diff.save()
        new_diff.check_conflicts()
        assert getattr(new_diff, "conflicts", None) is None

    @pytest.mark.django_db
    def test_check_conflicts_without_conflicts(self, diff_time_var):
        diff_time_var[4].check_conflicts()
        assert getattr(diff_time_var[4], "conflicts", None) is None

    @pytest.mark.django_db
    def test__init__with_default_values(self, diffusions):
        for diffusion in diffusions:
            assert diffusion._initial == {
                "start": diffusion.start,
                "end": diffusion.end,
                "episode": diffusion.episode,
            }
